/*
  This software is distributed under GNU General Public License, ver. 2
  or higher (at your option), released by Free Software Foundation. 
  You should receive text of the GNU GPL along with this software. If not,
  you can find it at:
     http://sageshome.net/GPL.php
   or
     http://www.gnu.org/
     http://www.opensource.org/
  
  Copyright(C) Sergey A. Galin, 2002-2004. 
 
  Change Log:
  
  * Thu Nov 11 2004 Sergey Galin <sergey@sageshome.net> 
  - v. 1.4 - Windows compiltaion (MinGW), added -x paramater.
  - v. 1.3 - Changed signature to allow processing of Eudora's mailboxes.  
  - v. 1.2 - Added patch by Florian Feucht (fixed a bug of skipping first 2 chars 
    from "From:" in the first message).
  
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>

#if !defined(USE_MINGW)
  #include <unistd.h>  
#else
  #include <time.h>
#endif

#define VERSION "1.4"

// Signature which separates messages:
const char *const Separator="\nFrom ";
// Number of symbols to skip in signature to start of the message:
const int Rewind_Count=1; 

int Silent=0;

#define MAXPATH 65535

#define NSPRINTF if(!Silent) printf

char outExtension[MAXPATH]="mbx";

int OutFile(const char *mdir){
  
  #if !defined(USE_MINGW)
    char buf[MAXPATH];
    snprintf(buf, sizeof(buf), "%s/mbox2mdirXXXXXX", mdir);
    return mkstemp(buf);
  #else
  
    char buf[MAXPATH];
    static long count = 0;
    snprintf(buf, sizeof(buf), "%s\\mbox2mdir_%ld_%03ld%05u.%s", 
      mdir, long(time(NULL)), count++, (unsigned)rand(), outExtension);
    printf("Writing file: %s\n", buf);
    int ret = open(buf, O_CREAT|O_EXCL|O_WRONLY);
    if( ret == -1 ){
       fprintf(stderr, "Failed to create new file: %s\n", buf);
       exit(-1);
    }else
      return ret;
      
  #endif
}

long strpos(const char *haystack, const char *needle){
  size_t nl=strlen(needle);
  for(size_t i=0; haystack[i]!='\0'; i++)
    if(memcmp((haystack+i), needle, nl)==0) return i;
  return -1;
}

size_t strdistance(const char *haystack, const char *needle){
  size_t nl=strlen(needle);
  for(size_t i=0; haystack[i]!='\0'; i++)
    if(memcmp((haystack+i), needle, nl)==0) return i;
  return strlen(haystack);
}

int FlushMail(const char *data, size_t sz, const char *dir){
  int f;
  if((f=OutFile(dir))<0){ perror("Unable to create output file"); return 1; }
  write(f, data, sz);
  #if !defined(USE_MINGW)
    fchmod(f, 0666);
  #endif
  close(f);
  return 0;
}

int Process(const char *ifile, const char *odir, int truncate){
  int inf;
  size_t sz;
  struct stat buf;  
  char *data, *seek;
  int count=0;
  
  inf=open(ifile, O_RDWR);
  NSPRINTF("Processing %s to %s...\n", ifile, odir);
  if(inf<0){ perror("Unable to open input file"); return 1; }
  
  #if !defined(USE_MINGW)
    if(flock(inf, LOCK_EX)==-1){ perror("Unable to lock input file"); return 1; }
  #endif
  
  fstat(inf, &buf);
  sz=buf.st_size;
  if(sz>0){
    data=(char *)malloc(sz+1);
    if(data==NULL){ perror("Cannot allocate buffer"); return 1; }  
    read(inf, data, sz);
    
    data[sz]=0;
    
    seek=data;
    
//size_t total=0;    
    do{

      size_t mailsz=strdistance(seek+1, Separator)+1;
//      printf("MAIL %d\n", mailsz);    
      // Added patch by Florian Feucht (fixed a bug of skipping first 2 chars 
      // from "From:" in the first message
      if (count==0) {
        // for the first time, no \n\n before mail starts.
        if(FlushMail(seek, mailsz, odir)){ 
          perror("Error processing mail, terminating"); 
	  close(inf);
	  return 1; 
        };
      }else{
        if(FlushMail(seek+Rewind_Count, mailsz, odir)){ 
          perror("Error processing mail, terminating"); 
	  close(inf);
	  return 1; 
        };
      }
      
	
      count++;
      
//total+=mailsz;      
      
      seek++;
    }while((seek=strstr(seek, Separator))!=NULL);
//printf("Total: %d\n", total);    
  
    free(data);
  }
  #if !defined(USE_MINGW)
    if(truncate)
      if(ftruncate(inf, 0)<0) 
        perror("Unable to truncate input file"); 
      else
        NSPRINTF("Input Mailbox truncated.");
  #endif
  close(inf);
  NSPRINTF("Processing %s to %s successful, %d messages (%d bytes) processed.\n", ifile, odir, count, sz);  
  return 0;
}

int FindParam(const char *par, int argc, char *argv[]){
  int parlen=strlen(par);
  for(int i=0; i<argc; i++)
    if(!strncmp(argv[i], par, parlen)) return i;    
  return -1;    
}

int main(int argc, char *argv[]){
  //const char *ifile="Mailbox", *odir="/home/sergey/Maildir/cur";
  const char *ifile, *odir;
  if(argc<3){
    printf("Mailbox to Maildir Converter, Version "VERSION" built: "__DATE__" Copyright(C) Sergey A. Galin.\n"
	   "This program is distibuted under GNU General Public License v.2 or later.\n"
	   "Homepage: http://sageshome.net/oss/\n"
	   "\n"
           "USAGE:   mbox2mdir <input mailbox file> <output maildir> [options]\n"
           "OPTIONS:\n"
	   #if !defined(USE_MINGW)
  	     "         -t      - Truncate input Maildir file after successful conversion.\n"
	   #else
	     "         -t      - Ignored in this MinGW build at the moment.\n"
	   #endif
	   "         -s      - Silent mode (error reporting only).\n"
	   
	   #if defined(USE_MINGW)
	   "         -e<EXT> - use EXT as filename extension.\n"
	   #endif 
	   
	   "EXAMPLE: $ mbox2mdir /var/qmail/alias/Mailbox /home/john/Maildir/cur -s -eeml\n");
  }else{
    int truncate=0, x;
    
    if(FindParam("-t", argc-3, argv+3)>=0) truncate=1;
    if(FindParam("-s", argc-3, argv+3)>=0) Silent=1;
    if((x=FindParam("-e", argc-3, argv+3))>=0){ 
      strncpy(outExtension, argv[3+x]+2, sizeof(outExtension)-1);
      //printf("Out extension: %s\n", outExtension);
    }
    ifile=argv[1];
    odir=argv[2];
    return Process(ifile, odir, truncate);    
  }
}
